#include "nBodySim.hpp"
#include <omp.h>
#include <cmath>
#include <string>
#include <fstream>
#include <iomanip>

nBodySim::nBodySim(std::string& datafile) {
    // reads in data from datafile
    std::string dummy;
    std::ifstream file(datafile);
    file >> dummy >> dummy >> N_;

    masses_ = new double[N_];
    for (unsigned i=0; i < N_; ++i) {
        file >> masses_[i];
    }

    positions_ = new double[3*N_];
    for (unsigned i=0; i < 3; ++i) {
        for (unsigned j=0; j < N_; ++j) {
            file >> positions_[i+3*j];
        }
    }
    
    velocities_ = new double[3*N_];
    for (unsigned i=0; i < 3; ++i) {
        for (unsigned j=0; j < N_; ++j) {
            file >> velocities_[i+3*j];
        }
    }

    softening_ = new double[N_];
    for (unsigned i=0; i < N_; ++i) {
        file >> softening_[i];
    }
    
    potential_ = new double[N_];
    for (unsigned i=0; i < N_; ++i) {
        file >> potential_[i];
    }

    forces_ = new double[3*N_];
    for (unsigned i=0; i < 3*N_; ++i) {
        forces_[i] = 0.0;
    }
}

nBodySim::~nBodySim() {
    delete[] masses_;
    delete[] positions_;
    delete[] velocities_;
    delete[] softening_;
    delete[] potential_;
    delete[] forces_;
}

void nBodySim::calculateForces() {
    #pragma omp parallel for
    for (unsigned i=0; i < 3*N_; ++i) {
        forces_[i] = 0.0;
    }
    // loop over each pair of particles and calculate the force between them
    #pragma omp parallel for collapse(2)
    for (unsigned i = 0; i < N_; i++) {
        for (unsigned j = 0; j < N_; j++) {
            if (i == j) continue;
            // calculate distance between particles
            double dx = positions_[3*j]   - positions_[3*i];
            double dy = positions_[3*j+1] - positions_[3*i+1];
            double dz = positions_[3*j+2] - positions_[3*i+2];
            double r2 = dx*dx + dy*dy + dz*dz;
            double r = std::sqrt(r2);
            dx /= r;
            dy /= r;
            dz /= r;
            // calculate force
            double mi = masses_[i];
            double mj = masses_[j];
            double s = softening_[i];
            forces_[3*i]   += mi*mj * dx / (r2 + s*s);
            forces_[3*i+1] += mi*mj * dy / (r2 + s*s);
            forces_[3*i+2] += mi*mj * dz / (r2 + s*s);
        }
    }
}

void nBodySim::write2file(const double* array, std::string filename, unsigned dim) const {
    // writes array to file, with N rows and dim columns
    std::ofstream file(filename);
    for (unsigned i = 0; i < N_; i++) {
        for (unsigned j = 0; j < dim; j++) {
            file << std::right << std::setw(16) << array[dim*i+j] << " ";
        }
        file << std::endl;
    }
}